(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Action_failed of string
exception Unresolved_promises



module Job = Value_types.Sum3
  (Rule.Target)
    (* For leaf targets. *)
  (Rule.Rule.Specialized(Rule.Targets))
    (* For normal rules. *)
  (Rule.Halfrule)
    (* For scope rules. *)

module Marking = Marking.Formula
  (Rule.Target)
  (Rule.TargetSet)
  (Job)

module Fold_store = Rulebase.Fold_store
  (Value_types.Unit)
  (Value_types.Unit)
  (Value_types.Unit)

module State_Control = Value_types.Tuple4
  (Marking)
  (Rule.Targets)
    (* The set of so far unfulfilled promises. *)
  (Rule.Targets)
    (* The set of (so far unfulfilled) promises
       which are needed for the current goals. *)
  (Rule.HalfruleSet)
    (* The set of rules (halfrules, because these are sufficient to identify
       rules) which have already been executed. *)

module State_TreeGeneration = Value_types.Pair
  (Oinmfile.Scope_Base)
    (* Database of flat scopes from the old tree. *)
  (Oinmfile.Scope_Base)
    (* Database of flat scopes already generated. *)

module State_Others = Debugable_types.Triple
  (Fold_store)
  (Hash.Hash_cache)
  (Uptodate.Global_Status)

module State = Debugable_types.Tuple4
  (State_Control)
  (State_TreeGeneration)
  (Rulebase.Rulebase)
  (State_Others)

type state = State.t


let handle_flat flatscope ?old_store
    old_control treegen old_rulebase hash_cache global  =
  let old_marking,old_promises,old_needed,executed = old_control  in

  let rulebase = Rulebase.merge_bases old_rulebase
    (Rulebase.base_of_scope flatscope)  in
  let _, subscoperules, new_promises = Oinmfile.split flatscope  in
  let promises = Rule.TargetSet.filter
    (fun target -> not (Rulebase.has_rule rulebase target))
    (Rule.TargetSet.union old_promises new_promises)  in
  let inter_needed,now_fulfilled = Rule.TargetSet.partition
    (fun target -> Rule.TargetSet.mem target promises)
    old_needed  in
  let new_subscopes = List.filter
    (fun halfrule -> not (Rule.HalfruleSet.mem halfrule executed))
    subscoperules  in
  let inter_marking,sources = List.fold_left
    (fun (marking,sources) -> fun halfrule ->
      Marking.add (Sums.S33 halfrule) halfrule.Rule.sources marking,
      Rule.TargetSet.union sources halfrule.Rule.sources)
    (old_marking,Rule.TargetSet.empty)
    new_subscopes  in
  let (marking,needed),store = Rulebase.fold_rulebase
    (fun (marking,needed) -> fun leaf -> if Rule.TargetSet.mem leaf promises
      then ((),(marking, Rule.TargetSet.add leaf needed))
      else ((),
        (Marking.add (Sums.S31 leaf) (Rule.TargetSet.empty) marking,
        needed)))
    (fun _ -> fun (marking,needed) -> fun _ -> (),(marking,needed))
    (fun (marking,needed) -> fun _ -> fun rule -> Rulebase.Enter (
      (),(), (Marking.add
          (Sums.S32 rule)
          rule.Rule.pre.Rule.sources
          marking,
        needed)))
    (fun _ -> fun _ -> fun (marking,needed) -> fun _ -> fun _ ->
      ((),(marking,needed)))
    (fun _ -> fun (marking,needed) -> fun _ -> fun _ -> ((),(marking,needed)))
    (fun _ -> fun _ -> fun (marking,needed) -> fun _ -> fun _ ->
      ((),(marking,needed)))
    ?store:old_store
    rulebase
    (inter_marking,inter_needed)
    (Rule.TargetSet.union sources now_fulfilled)  in

  let control = marking,promises,needed,executed  in
  let others = store,hash_cache,global  in
  control,treegen,rulebase,others


let init new_root old_tree hash_cache global targets =
  let marking = Marking.empty  in
  let promises = targets  in
  let needed = targets  in
  let executed = Rule.HalfruleSet.empty  in
  let control = marking,promises,needed,executed  in

  let old_scopes = Oinmfile.flatten old_tree  in
  let new_scopes = new_root, Rule.HalfruleMap.empty  in
  let treegen = old_scopes,new_scopes  in

  let rulebase = Rulebase.empty  in
  handle_flat new_root control treegen rulebase hash_cache global


let get_job state =
  let control,treegen,rulebase,others = state  in
  let marking,promises,needed,executed = control  in

  let job,marking = Marking.retrieve_positive marking  in

  let control = marking,promises,needed,executed  in
  let state = control,treegen,rulebase,others  in
  job,state


let consider_rule (execution : Action.Actions.t -> 'a) (default : 'a option)
    rule hash_cache global =
  (* Executes the rule ... unless it is still up to date and default!=None. *)

  let execute,source_hashes = Uptodate.pre_rule hash_cache global rule  in
  match execute,default  with
  | true,_ | _,None ->
    let result = execution rule.Rule.pre.Rule.actions  in
      (* It is important, that the actions are executed
         _before_ the new hashes are calculated. *)
    let new_drain_hashes =
      Uptodate.post_rule hash_cache global rule source_hashes  in
    result,new_drain_hashes
  | false,Some d -> d,global


let do_job_leaf leaf state =
  let control,treegen,rulebase,others = state  in
  let marking,promises,needed,executed = control  in

  let marking = Marking.set leaf marking  in

  let control = marking,promises,needed,executed  in
  let state = control,treegen,rulebase,others  in
  state


let do_job_files_rule rule state =
  let control,treegen,rulebase,others = state  in
  let marking,promises,needed,executed = control  in
  let store,hash_cache,global = others  in

  let success,global = consider_rule Action.prompt_exec_actions (Some true)
    (Rule.replace_drain rule (Rule.Files rule.Rule.drain))
    hash_cache  global  in
  let marking = if success
    then Rule.TargetSet.fold Marking.set rule.Rule.drain marking
    else raise (Action_failed "") (* todo *)  in
  let executed = Rule.HalfruleSet.add rule.Rule.pre executed  in

  let others = store,hash_cache,global  in
  let control = marking,promises,needed,executed  in
  let state = control,treegen,rulebase,others  in
  state


let do_job_scope_rule halfrule state =
  let control,treegen,rulebase,others = state  in
  let marking,promises,needed,executed = control  in
  let old_scopes,new_scopes = treegen  in
  let _,old_nonroot = old_scopes  in
  let new_root,new_nonroot = new_scopes  in
  let store,hash_cache,global = others  in

  let execution actions = match Action.prompt_exec_read actions  with
    | None -> raise (Action_failed "") (* todo *)
    | Some text -> Oinmfile.assume_flat (Oinmfile.merge_generated
      (Oinmfile_parser.oinmfile Oinmfile_lexer.oinmfile
        (Lexing.from_string text)))  in
  let scope,global = consider_rule execution
    (Rule.HalfruleMap.find_optional halfrule old_nonroot)
    {Rule.pre=halfrule; Rule.drain=Rule.Scope ()}
    hash_cache  global  in
  let new_nonroot = Rule.HalfruleMap.add_new halfrule scope new_nonroot  in
  let executed = Rule.HalfruleSet.add halfrule executed  in

  let new_scopes = new_root,new_nonroot  in
  let treegen = old_scopes,new_scopes  in
  let control = marking,promises,needed,executed  in
  handle_flat scope ~old_store:store control treegen rulebase hash_cache global


let do_job job = match job with
  | Sums.S31 target -> do_job_leaf target
  | Sums.S32 rule -> do_job_files_rule rule
  | Sums.S33 halfrule -> do_job_scope_rule halfrule


let run state =
  let rec do_it state =
    let job,state = get_job state  in
    match job with
    | None -> state
    | Some job -> do_it (do_job job state)  in
  let state = do_it state  in
  let _,treegen,rule_base,others = state  in
  let _,new_scopes = treegen  in
  let _,_,global = others  in
  (* todo: check whether all is well *)
  Oinmfile.combine new_scopes, global, rule_base

