(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Not_translatable



module type Type = sig

  type t

  val debug : t -> string

end



module type Iso = sig

  type from
  type to_

  val forth : from -> to_
  val back : to_ -> from
    (* If back is partial (that is from is isomorphic to a subtype of to_),
       then Not_translatable can be raised. *)

end

module Reduce(To : Type)(Iso : Iso with type to_ = To.t) = struct

  type t = Iso.from

  let debug f = "(after translation:) "^(To.debug (Iso.forth f))

end



module Unit = struct

  type t = unit

  let debug () = "()"

end

module Bool = struct

  type t = bool

  let debug b = if b then "true" else "false"

end

module Char = struct

  type t = char

  let debug c = "'" ^ Char.escaped c ^ "'"

end

module String = struct

  type t = string

  let debug s = s

end

module Int = struct

  type t = int

  let debug = string_of_int

end

module Float = struct

  type t = float

  let debug = string_of_float

end



module Option(T : Type) = struct

  type t = T.t option

  let debug o = match o with
  | None -> "-"
  | Some a -> "("^(T.debug a)^")"

end

module Ref(T : Type) = struct

  type t = T.t ref

  let debug r = T.debug !r

end



module Pair(Left : Type)(Right : Type) = struct

  type t = Left.t * Right.t

  let debug (l,r) = "("^(Left.debug l)^","^(Right.debug r)^")"

end

module Triple(T1 : Type)(T2 : Type)(T3 : Type) = struct

  type t = T1.t * T2.t * T3.t

  let debug (a1,a2,a3) =
    "("^(T1.debug a1)^","^(T2.debug a2)^","^(T3.debug a3)^")"

end

module Tuple4(T1 : Type)(T2 : Type)(T3 : Type)(T4 : Type) = struct

  type t = T1.t * T2.t * T3.t * T4.t

  let debug (a1,a2,a3,a4) =
    "("^(T1.debug a1)^","^(T2.debug a2)^","^(T3.debug a3)^","^(T4.debug a4)^")"

end

module Tuple5(T1 : Type)(T2 : Type)(T3 : Type)(T4 : Type)(T5 : Type) = struct

  type t = T1.t * T2.t * T3.t * T4.t * T5.t

  let debug (a1,a2,a3,a4,a5) =
    "("^(T1.debug a1)^","^(T2.debug a2)^","^(T3.debug a3)^","^
    (T4.debug a4)^","^(T5.debug a5)^")"

end



module Sum(Left : Type)(Right : Type) = struct

  type t = (Left.t,Right.t) Sums.sum

  let debug s = match s with
  | Sums.L l -> "Left "^(Left.debug l)
  | Sums.R r -> "Right "^(Right.debug r)

end

module Sum3(T1:Type)(T2:Type)(T3:Type) = struct

  type t = (T1.t,T2.t,T3.t) Sums.sum3

  let debug s = match s with
  | Sums.S31 v -> "Variant 1 "^(T1.debug v)
  | Sums.S32 v -> "Variant 2 "^(T2.debug v)
  | Sums.S33 v -> "Variant 3 "^(T3.debug v)

end



module List(Elem : Type) = struct

  type t = Elem.t list

  let debug l =
    (List.fold_left
      (fun d -> fun e -> d^(Elem.debug e)^"; ")
      "[" l)
    ^"]"

end



module Array(Elem : Type) = struct

  type t = Elem.t array

  let debug a =
    (Array.fold_left
      (fun d -> fun e -> d^(Elem.debug e)^"; ")
      "[" a)
    ^"]"

end

