(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Escape_at_end
exception Unknown_escape_code
exception Inconsistent_specification

let max_char_code = 127

type escape_action = Keep | Escape of char
type unescape_action = Delete | Error | Replace of char

type escape_spec = char * escape_action array * unescape_action array

let escape_spec c direct indirect deletes =
  let replaces = (List.map (fun c -> c,c) direct) @ indirect  in
  let escape = Array.init (max_char_code+1) (fun n -> Keep)  in
  let unescape = Array.init (max_char_code+1) (fun n -> Error)  in
  List.iter (fun (real,code) ->
      let rn,cn = int_of_char real, int_of_char code  in
      (if rn>max_char_code || cn>max_char_code
        then raise Inconsistent_specification);
      (if escape.(rn) <> Keep || unescape.(cn) <> Error
        then raise Inconsistent_specification);
      escape.(rn) <- Escape code;
      unescape.(cn) <- Replace real)
    replaces;
  List.iter (fun code ->
      let cn = int_of_char code  in
      (if cn>max_char_code  then raise Inconsistent_specification);
      (if unescape.(cn) <> Error  then raise Inconsistent_specification);
      unescape.(cn) <- Delete)
    deletes;
  c,escape,unescape


let escape_string (esc_char,spec,_) string =
  let out = String.create (2 * String.length string)  in
  let out_index = ref 0  in
  for in_index = 0 to String.length string - 1 do
    let c = string.[in_index]  in
    let cn = int_of_char c  in
    let action = if cn>max_char_code  then Keep  else spec.(cn)  in
    match action with
    | Keep -> (
      out.[!out_index] <- c;
      out_index := !out_index+1)
    | Escape c' -> (
      out.[!out_index] <- esc_char;
      out.[!out_index+1] <- c';
      out_index := !out_index+2)
  done;
  String.sub out 0 !out_index

let unescape_string (esc_char,_,spec) string =
  let in_len = String.length string  in
  let out = String.create in_len  in
  let in_index = ref 0  in
  let out_index = ref 0  in
  while !in_index < in_len do
    let c = string.[!in_index]  in
    if c=esc_char
    then (
      in_index := !in_index+1;
      if !in_index = in_len  then raise Escape_at_end;
      let cn = int_of_char string.[!in_index]  in
      let action = if cn>max_char_code  then Error  else spec.(cn)  in
      match action with
      | Delete -> ()
      | Error -> raise Unknown_escape_code
      | Replace c' -> out.[!out_index] <- c' )
    else out.[!out_index] <- c;
    in_index := !in_index+1;
    out_index := !out_index+1;
  done;
  String.sub out 0 !out_index

let unescape_lexeme (esc_char,_,spec) lexbuf =
  let in_len = Lexing.lexeme_end lexbuf - Lexing.lexeme_start lexbuf  in
  let out = String.create in_len  in
  let in_index = ref 0  in
  let out_index = ref 0  in
  while !in_index < in_len do
    let c = Lexing.lexeme_char lexbuf !in_index  in
    if c=esc_char
    then (
      in_index := !in_index+1;
      if !in_index = in_len  then raise Escape_at_end;
      let cn = int_of_char (Lexing.lexeme_char lexbuf !in_index)  in
      let action = if cn>max_char_code  then Error  else spec.(cn)  in
      match action with
      | Delete -> ()
      | Error -> raise Unknown_escape_code
      | Replace c' -> out.[!out_index] <- c' )
    else out.[!out_index] <- c;
    in_index := !in_index+1;
    out_index := !out_index+1;
  done;
  String.sub out 0 !out_index



let line_escape_spec = escape_spec '%' ['%'] ['\n','n'] []



let shell_escape string =
  (* If the implementation is changed, for example to optimize the length of
     the escaped string, then it should not be forgotten that the empty
     string needs to be quoted. *)
  let out = String.make (2 + 4 * String.length string) '\''  in
  let out_index = ref 1  in
  for in_index = 0 to String.length string - 1 do
    let c = string.[in_index]  in
    if c='\''
    then (
      out.[!out_index+1] <- '\\';
      out_index := !out_index+4;
      )
    else (
      out.[!out_index] <- c;
      out_index := !out_index+1;
      );
  done;
  String.sub out 0 (!out_index+1)

