(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Not_flat


type
  node =
  | Rule of tree option Rule.drain_cases Rule.rule
  | Generated of Rule.TargetSet.t
and
  tree = node list

let fold_tree
    f_rule_files
    f_rule_enter_scope
    f_rule_exit_scope
    f_rule_scope_undone
    f_generated
    a  t =
  let rec foldtree a t = List.fold_left
    (fun a -> fun node -> match node with
      | Rule rule -> (match rule.Rule.drain with
        | Rule.Files files -> f_rule_files a
          (Rule.replace_drain rule files)
        | Rule.Scope None -> f_rule_scope_undone a rule.Rule.pre
        | Rule.Scope (Some scope) ->
          let skip,up = f_rule_enter_scope a rule.Rule.pre  in
          f_rule_exit_scope skip (foldtree up scope) rule.Rule.pre)
      | Generated targets -> f_generated a targets)
    a  t  in
  foldtree a t

let rec merge_generated tree = match tree with
  | Rule r :: t ->
    let t' = merge_generated t  in
    (match r.Rule.drain with
    | Rule.Files _ | Rule.Scope None -> Rule r :: t'
    | Rule.Scope (Some sub) ->
      Rule {r with Rule.drain = Rule.Scope (Some (merge_generated sub))} :: t')
  | Generated g1 :: Generated g2 :: t ->
    merge_generated (Generated (Rule.TargetSet.union g1 g2) :: t)
  | Generated g :: t -> Generated g :: merge_generated t
  | [] -> []

let escape_target = Escape.escape_string Oinmfile_escape.target_spec
let escape_action = Escape.escape_string Oinmfile_escape.action_spec

let output_tree out tree =
  let out_set prefix set =
    let last = Rule.TargetSet.fold
      (fun target -> fun sofar ->
        let name = escape_target target  in
        let next = sofar^" "^name  in
        if String.length next < 80
        then next
        else (
          output_string out (sofar^"\n");
          prefix^" "^name))
      set
      prefix  in
    output_string out (last^"\n")  in
  let out_halfrule prefix halfrule =
    out_set (prefix^"<") halfrule.Rule.sources;
    List.iter
      (fun action -> output_string out
        (prefix^"  "^(escape_action action)^"\n"))
      (Action.strings_of_actions halfrule.Rule.actions)  in
  let rec do_it indent tree = List.iter
    (fun node -> match node with
      | Rule rule -> (
        out_halfrule indent rule.Rule.pre;
        match rule.Rule.drain with
        | Rule.Files files -> out_set (indent^">") files
        | Rule.Scope None -> output_string out (indent^"{}\n")
        | Rule.Scope (Some sub) -> (
          output_string out (indent^"{\n");
          do_it (indent^"  ") sub;
          output_string out (indent^"}\n")))
      | Generated targets -> if not (Rule.TargetSet.is_empty targets)
        then out_set (indent^"%generated") targets)
    tree  in
  do_it "" tree



type
  flat_node =
  | FRule of unit Rule.drain_cases Rule.rule
  | FGenerated of Rule.TargetSet.t
and
  flat_scope = flat_node list

module FN_Iso = struct

  type from = flat_node
  type to_ =
    (unit Rule.drain_cases Rule.rule, Rule.TargetSet.t) Sums.sum

  let forth n = match n with
  | FRule r -> Sums.L r
  | FGenerated g -> Sums.R g

  let back s = match s with
  | Sums.L r -> FRule r
  | Sums.R g -> FGenerated g

end

module Flat_Node = Value_types.Reduce
  (Value_types.Sum
    (Rule.Rule.Specialized(Rule.Drain_cases.Specialized(Value_types.Unit)))
    (Rule.TargetSet))
  (FN_Iso)

module Flat_Scope = Value_types.List(Flat_Node)

let assume_flat = List.map
  (fun node -> match node with
  | Rule r -> FRule (Rule.replace_drain r (match r.Rule.drain with
    | Rule.Files fs -> Rule.Files fs
    | Rule.Scope None -> Rule.Scope ()
    | Rule.Scope (Some _) -> raise Not_flat))
  | Generated g -> FGenerated g)

let flat_to_tree = List.map
  (fun node -> match node with
  | FRule r -> Rule (Rule.replace_drain r (match r.Rule.drain with
    | Rule.Files fs -> Rule.Files fs
    | Rule.Scope () -> Rule.Scope None))
  | FGenerated g -> Generated g)

let split flat = List.fold_left
  (fun (files,subscopes,generated) -> fun fnode -> match fnode with
  | FRule r -> (match r.Rule.drain with
    | Rule.Files fs -> ((Rule.replace_drain r fs)::files,
      subscopes, generated)
    | Rule.Scope s -> (files, r.Rule.pre::subscopes, generated))
  | FGenerated g -> (files, subscopes, Rule.TargetSet.union g generated))
  ([],[],Rule.TargetSet.empty)
  flat


module Scope_Base = Value_types.Pair
  (Flat_Scope)
  (Rule.HalfruleMap.Specialized(Flat_Scope))

let flatten tree =
  let rec do_it scope map =
    List.fold_right
      (fun node -> fun (fscope,map) -> match node with
        | Rule r ->
          let drain,map = (match r.Rule.drain with
          | Rule.Files fs -> Rule.Files fs,map
          | Rule.Scope sub -> Rule.Scope (),(match sub with
            | None -> map
            | Some sub ->
              let fsubscope,map = do_it sub map  in
              Rule.HalfruleMap.add_new r.Rule.pre fsubscope map)) in
          (FRule (Rule.replace_drain r drain))::fscope,map
        | Generated g -> (FGenerated g)::fscope,map)
      scope
      ([],map)  in
  do_it tree Rule.HalfruleMap.empty

let combine scopebase =
  let root,nonroot = scopebase  in
  let expanded = ref Rule.HalfruleSet.empty  in
  let rec do_it scope = List.map
    (fun flat_node -> match flat_node  with
      | FRule rule -> Rule (match rule.Rule.drain with
        | Rule.Files targets -> Rule.replace_drain rule (Rule.Files targets)
        | Rule.Scope () -> if Rule.HalfruleSet.mem rule.Rule.pre !expanded
          then Rule.replace_drain rule (Rule.Scope None)
          else (
            let subscope = Rule.HalfruleMap.find rule.Rule.pre nonroot  in
            expanded := Rule.HalfruleSet.add rule.Rule.pre !expanded;
            Rule.replace_drain rule (Rule.Scope (Some (do_it subscope)))))
      | FGenerated targets -> Generated targets)
    scope  in
  do_it root

