(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Not_translatable
exception Malformed_input
exception Double_Definition

module List_original = List



module type Type = sig

  include Debugable_types.Type

  val compare : t -> t -> int
  val save : out_channel -> t -> unit
  val load : in_channel -> t

end

module type PType = sig

  type 'a pt

  module Specialized(A : Type) : Type with type t = A.t pt

end



module Reduce(To : Type)(Iso : Debugable_types.Iso with type to_ = To.t) =
  struct

  include Debugable_types.Reduce(To)(Iso)

  let compare f1 f2 = To.compare (Iso.forth f1) (Iso.forth f2)
  let save file f = To.save file (Iso.forth f)
  let load file = Iso.back (To.load file)

end



module Unit = struct

  include Debugable_types.Unit

  let compare () () = 0
  let save file () = ()
  let load file = ()

end

module Bool = struct

  include Debugable_types.Bool

  let compare b1 b2 = if b1 then if b2 then 0 else 1 else if b2 then -1 else 0
  let save file b = output_string file (if b then "+\n" else "-\n")
  let load file = try
    match input_line file with
    | "+" -> true
    | "-" -> false
    | _ -> raise Malformed_input
  with
  | End_of_file -> raise Malformed_input

end

module Char = struct

  include Debugable_types.Char

  let compare = compare
  let save file c = output_string file (string_of_int (Char.code c) ^"\n")
  let load file = try
    Char.chr (int_of_string (input_line file))
  with
  | End_of_file
  | Failure "int_of_string"
  | Invalid_argument "Char.chr" -> raise Malformed_input

end

module String = struct

  include Debugable_types.String

  let compare = compare
  let save file s =
    output_string file ((Escape.escape_string Escape.line_escape_spec s)^"\n")
  let load file = try
    Escape.unescape_string Escape.line_escape_spec (input_line file)
  with
  | End_of_file
  | Escape.Escape_at_end | Escape.Unknown_escape_code -> raise Malformed_input

end

module Int = struct

  include Debugable_types.Int

   let compare = compare
   let save file i = output_string file ((string_of_int i)^"\n")
   let load file = try
     int_of_string (input_line file)
   with
   | End_of_file
   | Failure "int_of_string" -> raise Malformed_input

end

module Float = struct
  include Debugable_types.Float

  let compare = compare
  let save file f = output_string file
    (Int64.to_string (Int64.bits_of_float f) ^ "\n")
  let load file = try
    Int64.float_of_bits (Int64.of_string (input_line file))
  with
  | End_of_file | Failure "int_of_string" -> raise Malformed_input
end



module Option(T : Type) = struct

  include Debugable_types.Option(T)

  let compare o1 o2 = match o1,o2 with
  | None,None -> 0
  | None,Some _ -> -1
  | Some _, None -> 1
  | Some a1, Some a2 -> T.compare a1 a2

  let save file o = match o with
  | None -> Bool.save file false
  | Some a -> (
    Bool.save file true;
    T.save file a)

  let load file = match Bool.load file with
  | false -> None
  | true -> Some (T.load file)

end



module Pair(Left : Type)(Right : Type) = struct

  include Debugable_types.Pair(Left)(Right)

  let compare (l1,r1) (l2,r2) =
    let cl = Left.compare l1 l2  in
    if cl=0  then Right.compare r1 r2  else cl

  let save file (l,r) = Left.save file l; Right.save file r

  let load file =
    let l = Left.load file  in
    let r = Right.load file  in
    (l,r)

end

module Triple(T1 : Type)(T2 : Type)(T3 : Type) = struct

  include Debugable_types.Triple(T1)(T2)(T3)

  let compare (a1,a2,a3) (b1,b2,b3) =
    let c1 = T1.compare a1 b1  in
    if c1=0
    then let c2 = T2.compare a2 b2  in
      if c2=0
      then T3.compare a3 b3
      else c2
    else c1

  let save file (a1,a2,a3) = T1.save file a1; T2.save file a2; T3.save file a3

  let load file =
    let a1 = T1.load file  in
    let a2 = T2.load file  in
    let a3 = T3.load file  in
    (a1,a2,a3)

end

module Tuple4(T1 : Type)(T2 : Type)(T3 : Type)(T4 : Type) = struct

  include Debugable_types.Tuple4(T1)(T2)(T3)(T4)

  let compare (a1,a2,a3,a4) (b1,b2,b3,b4) =
    let c1 = T1.compare a1 b1  in
    if c1=0
    then let c2 = T2.compare a2 b2  in
      if c2=0
      then let c3 = T3.compare a3 b3  in
        if c3=0
        then T4.compare a4 b4
        else c3
      else c2
    else c1

  let save file (a1,a2,a3,a4) =
    T1.save file a1; T2.save file a2; T3.save file a3; T4.save file a4

  let load file =
    let a1 = T1.load file  in
    let a2 = T2.load file  in
    let a3 = T3.load file  in
    let a4 = T4.load file  in
    (a1,a2,a3,a4)

end

module Tuple5(T1 : Type)(T2 : Type)(T3 : Type)(T4 : Type)(T5 : Type) = struct

  include Debugable_types.Tuple5(T1)(T2)(T3)(T4)(T5)

  let compare (a1,a2,a3,a4,a5) (b1,b2,b3,b4,b5) =
    let c1 = T1.compare a1 b1  in
    if c1=0
    then let c2 = T2.compare a2 b2  in
      if c2=0
      then let c3 = T3.compare a3 b3  in
        if c3=0
        then let c4 = T4.compare a4 b4  in
          if c4=0
          then T5.compare a5 b5
          else c4
        else c3
      else c2
    else c1

  let save file (a1,a2,a3,a4,a5) =
    T1.save file a1; T2.save file a2; T3.save file a3;
    T4.save file a4; T5.save file a5

  let load file =
    let a1 = T1.load file  in
    let a2 = T2.load file  in
    let a3 = T3.load file  in
    let a4 = T4.load file  in
    let a5 = T5.load file  in
    (a1,a2,a3,a4,a5)

end



module Sum(Left : Type)(Right : Type) = struct

  include Debugable_types.Sum(Left)(Right)

  let compare s1 s2 = match s1,s2 with
  | Sums.L l, Sums.R r -> 1
  | Sums.R r, Sums.L l -> -1
  | Sums.L l1, Sums.L l2 -> Left.compare l1 l2
  | Sums.R r1, Sums.R r2 -> Right.compare r1 r2

  let save file s = match s with
  | Sums.L l -> (Bool.save file true; Left.save file l)
  | Sums.R r -> (Bool.save file false; Right.save file r)

  let load file = if Bool.load file
  then Sums.L (Left.load file)
  else Sums.R (Right.load file)

end

module Sum3(T1:Type)(T2:Type)(T3:Type) = struct

  include Debugable_types.Sum3(T1)(T2)(T3)

  let compare s1 s2 = match s1,s2 with
  | Sums.S31 v1, Sums.S31 v2 -> T1.compare v1 v2
  | Sums.S32 v1, Sums.S32 v2 -> T2.compare v1 v2
  | Sums.S33 v1, Sums.S33 v2 -> T3.compare v1 v2
  | Sums.S31 _, _ -> 1
  | Sums.S32 _, Sums.S33 _ -> 1
  | _ -> -1

  let save file s = match s with
  | Sums.S31 v -> (Int.save file 1; T1.save file v)
  | Sums.S32 v -> (Int.save file 2; T2.save file v)
  | Sums.S33 v -> (Int.save file 3; T3.save file v)

  let load file = match Int.load file with
  | 1 -> Sums.S31 (T1.load file)
  | 2 -> Sums.S32 (T2.load file)
  | 3 -> Sums.S33 (T3.load file)
  | _ -> raise Malformed_input

end



module List(Elem : Type) = struct

  include Debugable_types.List(Elem)

  let rec compare l1 l2 = match l1,l2 with
  | [], [] -> 0
  | [], h2::t2 -> -1
  | h1::t1, [] -> 1
  | h1::t1, h2::t2 ->
    let hc = Elem.compare h1 h2  in
    if hc=0  then compare t1 t2  else hc

  let save file l =
    Int.save file (List.length l);
    List.iter (Elem.save file) l

  let load file =
    let rec doit i = if i=0
    then []
    else
      let e = (Elem.load file)  in
      e::(doit (i-1))  in
    doit (Int.load file)

end



module Array(Elem : Type) = struct

  include Debugable_types.Array(Elem)

  let compare a1 a2 =
    let l1 = Array.length a1  in
    let l2 = Array.length a2  in
    let cl = Int.compare l1 l2  in
    if cl=0
    then
      let rec doit i = if i=l1
      then 0
      else let c = Elem.compare a1.(i) a2.(i)  in
        if c=0  then doit (i+1)  else c  in
      doit 0
    else cl

  let save file a =
    Int.save file (Array.length a);
    Array.iter (Elem.save file) a

  let load file =
      (* This is more complicated than it would be, if Array.init guaranteed
         to initialize in ascending order. *)
    let len = Int.load file  in
    if len=0
    then [| |]
    else (
      let e0 = Elem.load file  in
      let a = Array.make len e0  in
      for i=1 to len-1 do
        a.(i) <- Elem.load file
      done;
      a)

end



module type SetT = sig

  include Type

  module Elem : Type

  val empty : t
  val is_empty : t -> bool
  val mem : Elem.t -> t -> bool
  val add : Elem.t -> t -> t
  val remove : Elem.t -> t -> t
  val singleton : Elem.t -> t
  val union : t -> t -> t
  val diff : t -> t -> t
  val min_elt : t -> Elem.t
  val max_elt : t -> Elem.t
  val choose : t -> Elem.t
  val fold : (Elem.t -> 'b -> 'b) -> t -> 'b -> 'b
  val iter : (Elem.t -> unit) -> t -> unit
  val filter : (Elem.t -> bool) -> t -> t
  val partition : (Elem.t -> bool) -> t -> t * t
  val from_list : Elem.t list -> t

end

module Set(Elem : Type) = struct

  module Elem = Elem
  module S = Set.Make(Elem)

  type t = S.t

  let compare = S.compare

  let debug s =
    (S.fold
      (fun e -> fun d -> d^(Elem.debug e)^",")
      s
      "{")
    ^"}"

  let save file s =
    Int.save file (S.cardinal s);
    S.iter (Elem.save file) s

  let load file =
    let size = Int.load file  in
    let s = ref S.empty in
    for i=1 to size do
      s := S.add (Elem.load file) !s
    done;
    !s

  let empty = S.empty
  let is_empty = S.is_empty
  let mem = S.mem
  let add = S.add
  let remove = S.remove
  let singleton = S.singleton
  let union = S.union
  let diff = S.diff
  let min_elt = S.min_elt
  let max_elt = S.max_elt
  let choose = S.choose
  let fold = S.fold
  let iter = S.iter
  let filter = S.filter
  let partition = S.partition

  let from_list = List_original.fold_left
    (fun s -> fun e -> add e s)
    empty

end



module type MapT = sig

  include PType

  module Key : Type

  val empty : 'a pt
  val is_empty : 'a pt -> bool
  val mem : Key.t -> 'a pt -> bool
  val add : Key.t -> 'a -> 'a pt -> 'a pt
  val add_new : Key.t -> 'a -> 'a pt -> 'a pt
  val remove : Key.t -> 'a pt -> 'a pt
  val find : Key.t -> 'a pt -> 'a
  val find_optional : Key.t -> 'a pt -> 'a option
  val find_default : 'a -> Key.t -> 'a pt -> 'a
  val fold : (Key.t -> 'a -> 'b -> 'b) -> 'a pt -> 'b -> 'b

end

module Map(Key : Type) = struct

  module Key = Key
  module M = Map.Make(Key)

  type 'a pt = 'a M.t

  module Specialized(Elem : Type) = struct

    type t = Elem.t M.t

    let compare = M.compare Elem.compare

    let debug m =
      (M.fold
        (fun k -> fun e -> fun d -> d^(Key.debug k)^" -> "^(Elem.debug e)^"; ")
        m  "[[ ")
      ^ "]]\n"

    let save file m =
      let size = M.fold (fun k -> fun e -> fun n -> n+1) m 0  in
      Int.save file size;
      M.iter
        (fun k -> fun e -> Key.save file k; Elem.save file e)
        m

    let load file =
      let size = Int.load file  in
      let m = ref M.empty in
      for i=1 to size do
        let k = Key.load file  in
        let e = Elem.load file  in
        m := M.add k e !m
      done;
      !m

  end

  let empty = M.empty
  let is_empty = M.is_empty
  let mem = M.mem
  let add = M.add

  let add_new key a m = if mem key m
  then raise Double_Definition
  else add key a m

  let remove = M.remove
  let find = M.find

  let find_optional key map = try
      Some (M.find key map)
    with Not_found -> None

  let find_default default key map = match find_optional key map  with
  | None -> default
  | Some e -> e

  let fold = M.fold

end

