(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Not_translatable
exception Malformed_input
exception Double_Definition

module type Type = sig

  include Debugable_types.Type

  val compare : t -> t -> int
  val save : out_channel -> t -> unit
  val load : in_channel -> t
    (* save writes a value to the file, so that it can be retrieved with load.
       save and load should not assume that the channel consists of only this
       one value.
       load may raise Malformed_input. *)

end

module type PType = sig

  type 'a pt

  module Specialized(A : Type) : Type with type t = A.t pt

end

module Unit : Type with type t = unit
module Bool : Type with type t = bool
module Char : Type with type t = char
module String : Type with type t = string
module Int : Type with type t = int
module Float : Type with type t = float

module Reduce(To : Type)(Iso : Debugable_types.Iso with type to_ = To.t) :
  Type with type t=Iso.from

module Option(T : Type) : Type with type t=T.t option

(* In the following, compare is big endian lexical comparison. *)
module Pair(Left : Type)(Right : Type) : Type with type t = Left.t * Right.t
module Triple(T1 : Type)(T2 : Type)(T3 : Type) :
  Type with type t = T1.t * T2.t * T3.t
module Tuple4(T1 : Type)(T2 : Type)(T3 : Type)(T4 : Type) :
  Type with type t = T1.t * T2.t * T3.t * T4.t
module Tuple5(T1 : Type)(T2 : Type)(T3 : Type)(T4 : Type)(T5 : Type) :
  Type with type t = T1.t * T2.t * T3.t * T4.t * T5.t

module Sum(Left : Type)(Right : Type) : Type
  with type t = (Left.t,Right.t) Sums.sum
module Sum3(T1:Type)(T2:Type)(T3:Type) : Type
  with type t = (T1.t,T2.t,T3.t) Sums.sum3

module List(Elem : Type) : Type with type t=Elem.t list
module Array(Elem : Type) : Type with type t=Elem.t array

module type SetT = sig

  include Type

  module Elem : Type

  val empty : t
  val is_empty : t -> bool
  val mem : Elem.t -> t -> bool
  val add : Elem.t -> t -> t
  val remove : Elem.t -> t -> t
  val singleton : Elem.t -> t
  val union : t -> t -> t
  val diff : t -> t -> t
  val min_elt : t -> Elem.t
  val max_elt : t -> Elem.t
  val choose : t -> Elem.t
  val fold : (Elem.t -> 'b -> 'b) -> t -> 'b -> 'b
  val iter : (Elem.t -> unit) -> t -> unit
  val filter : (Elem.t -> bool) -> t -> t
  val partition : (Elem.t -> bool) -> t -> t * t
  val from_list : Elem.t list -> t

end
module Set(Elem : Type) : SetT with module Elem = Elem

module type MapT = sig

  include PType

  module Key : Type

  val empty : 'a pt
  val is_empty : 'a pt -> bool
  val mem : Key.t -> 'a pt -> bool
  val add : Key.t -> 'a -> 'a pt -> 'a pt
  val add_new : Key.t -> 'a -> 'a pt -> 'a pt
    (* Raises Double_Definition, if the key was already bound. *)
  val remove : Key.t -> 'a pt -> 'a pt
  val find : Key.t -> 'a pt -> 'a
  val find_optional : Key.t -> 'a pt -> 'a option
  val find_default : 'a -> Key.t -> 'a pt -> 'a
  val fold : (Key.t -> 'a -> 'b -> 'b) -> 'a pt -> 'b -> 'b

end
module Map(Key : Type) : MapT with module Key = Key

