(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

module Formula
    (Variable : Value_types.Type)
    (VariableSet : Value_types.SetT with module Elem = Variable)
    (Clause : Value_types.Type) = struct

  module VarSet = Value_types.Set(Variable)
    (* Although VariableSet has a suitable interface,
       we do not trust it to be the best implementation. *)
  module VarMap = Value_types.Map(Variable)
  module ClauseSet = Value_types.Set(Clause)
  module ClauseMap = Value_types.Map(Clause)

  include Value_types.Tuple4
    (VarMap.Specialized(ClauseSet))
      (* Tracks the clauses which have a certain variable as precondition. *)
    (ClauseMap.Specialized(VarSet))
      (* Tracks the unmet preconditions. *)
    (VarSet)
      (* Tracks the met variables. *)
    (ClauseSet)
      (* Tracks the clauses without unmet preconditions. *)

  let empty = VarMap.empty, ClauseMap.empty, VarSet.empty, ClauseSet.empty

  let is_empty (occurences,unmet,met,fullymet) =
    (ClauseMap.is_empty unmet) && (ClauseSet.is_empty fullymet)

  let add clause preconditions (occurences,unmet,met,fullymet) =
    if ClauseMap.mem clause unmet
    then (occurences,unmet,met,fullymet)
    else
      let pres = VariableSet.fold VarSet.add preconditions VarSet.empty  in
      let pres = VarSet.diff pres met  in
      let occs = VarSet.fold
        (fun var -> fun occs ->
          let old_var_occs = VarMap.find_default
            ClauseSet.empty  var  occs  in
          VarMap.add var (ClauseSet.add clause old_var_occs) occs)
        pres
        occurences  in
      let unmet = ClauseMap.add clause pres unmet  in
      let fullymet = if VarSet.is_empty pres
      then ClauseSet.add clause fullymet
      else fullymet  in
      (occs,unmet,met,fullymet)

  let retrieve_positive formula =
    match formula with (occurences,unmet,met,fullymet) ->
    if ClauseSet.is_empty fullymet
    then None,formula
    else
      let clause = ClauseSet.choose fullymet  in
      Some clause, (occurences,unmet,met,ClauseSet.remove clause fullymet)

  let set variable (occurences,unmet,met,fullymet) =
    let met = VarSet.add variable met  in
    match VarMap.find_optional variable occurences  with
    | None -> (occurences,unmet,met,fullymet)
    | Some clauseset ->
      let unmet,fullymet = ClauseSet.fold
        (fun clause -> fun (unmet,fullymet) ->
          let pres = VarSet.remove variable (ClauseMap.find clause unmet)  in
          let unmet = ClauseMap.add clause pres unmet  in
          let fullymet = if VarSet.is_empty pres
          then ClauseSet.add clause fullymet
          else fullymet  in
          unmet,fullymet)
        clauseset
        (unmet,fullymet)  in
      (occurences,unmet,met,fullymet)

end

