(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Multiple_Rules of string

module Rulebase = Debugable_types.Pair
  (Rule.TargetMap.Specialized(Value_types.Int))
    (* The ints are indices for the rule array below. *)
  (Debugable_types.Array(Rule.Rule.Specialized(Rule.Targets)))

type drain = Rule.Targets.t

let empty = Rule.TargetMap.empty, [| |]

let base_of_rules rules =
  let indices,num_rules = Array.fold_left
    (fun (indices,index) -> fun rule ->
      let indices = Rule.TargetSet.fold
        (fun target -> fun indices ->
          if Rule.TargetMap.mem target indices
          then raise (Multiple_Rules target)
          else Rule.TargetMap.add target index indices)
        rule.Rule.drain
        indices  in
      (indices,index+1))
    (Rule.TargetMap.empty,0)
    rules  in
  (indices,rules)

let merge_bases (indices1,rules1) (indices2,rules2) =
  base_of_rules (Array.append rules1 rules2)

let base_of_tree tree =
  let rules = Oinmfile.fold_tree
      (* 'walk = targets rule list
         'skip = unit *)
    (fun rules -> fun rule -> rule::rules)
    (fun walk -> fun _ -> (),walk)
    (fun () -> fun walk -> fun _ -> walk)
    (fun walk -> fun _ -> walk)
    (fun walk -> fun _ -> walk)
    []
    tree  in
  base_of_rules (Array.of_list rules)

let base_of_scope scope =
  let filerules,_,_ = Oinmfile.split scope  in
  base_of_rules (Array.of_list filerules)

let has_rule (indices,rules) target = Rule.TargetMap.mem target indices

let get_rule (indices,rules) target =
  let index = Rule.TargetMap.find target indices  in
  rules.(index)

let get_rule_optional (indices,rules) target =
  match Rule.TargetMap.find_optional target indices with
  | None -> None
  | Some index -> Some rules.(index)

let num_rules (_,rules) = Array.length rules


type 'a rulemap = 'a option array
  (* A rulemap for a rulebase b maps rules of b to values of type 'a
     via the indexing inherent in b. It should still be valid for the rulebase
     (merge_bases b b'). *)

let rulemap_init rulebase =
  Array.make (num_rules rulebase) None

let rulemap_extend rulemap rulebase =
  let map_len = Array.length rulemap  in
  let num_rules = num_rules rulebase  in
  if map_len >= num_rules
  then rulemap
  else Array.append rulemap (Array.make (num_rules - map_len) None)


type ('walk,'skip,'cycle,'confluence) subdigraph_options =
  | Enter of 'skip * 'cycle * 'walk
  | Ignore of 'cycle * 'confluence * 'walk

type ('leaf,'cycle,'confluence) fold_store =
  'leaf Rule.TargetMap.pt *
  'cycle option array *
  'confluence option array

module Fold_store
  (Leaf : Value_types.Type)
  (Cycle : Value_types.Type)
  (Confluence : Value_types.Type) =
  Debugable_types.Triple
    (Rule.TargetMap.Specialized(Leaf))
    (Debugable_types.Array(Value_types.Option(Cycle)))
    (Debugable_types.Array(Value_types.Option(Confluence)))


let fold_rulebase
    f_leaf
    f_revisit_leaf
    f_rule_start
    f_rule_end
    f_revisit_cycle
    f_revisit_confluence
    ?store
    rulebase  a  entry_points =
  let indices,rules = rulebase  in
  let leaf_store,cycle_store,confluence_store = match store with
  | None ->
    (Rule.TargetMap.empty,
    rulemap_init rulebase,
    rulemap_init rulebase)
  | Some (ls,cys,cos) ->
    (ls,
    rulemap_extend cys rulebase,
    rulemap_extend cos rulebase)  in
  let rec fold targets a leaf_store =
    Rule.TargetSet.fold
      (fun target -> fun (a,leaf_store) ->
        match Rule.TargetMap.find_optional target indices  with
        | None ->
          let leaf,a =
            (match Rule.TargetMap.find_optional target leaf_store  with
            | None -> f_leaf a target
            | Some leaf -> f_revisit_leaf leaf a target)  in
          a,
          Rule.TargetMap.add target leaf leaf_store
        | Some index ->
          let rule = rules.(index)  in
          (match cycle_store.(index) with
          | None -> (match f_rule_start a target rule with
	    | Enter (skip,cycle,a) -> (
              cycle_store.(index) <- Some cycle;
	      let a,leaf_store = fold
                rule.Rule.pre.Rule.sources  a  leaf_store  in
              match cycle_store.(index) with Some cycle ->
              let confluence,a = f_rule_end skip cycle a target rule in
              confluence_store.(index) <- Some confluence;
              (a,leaf_store))
	    | Ignore (cycle,confluence,a) -> (
              cycle_store.(index) <- Some cycle;
              confluence_store.(index) <- Some confluence;
              (a,leaf_store)))
          | Some cycle -> (match confluence_store.(index) with
            | None -> (
              let cycle,a = f_revisit_cycle cycle a target rule  in
              cycle_store.(index) <- Some cycle;
              (a,leaf_store))
            | Some confluence -> (
              let confluence,a = f_revisit_confluence
                cycle confluence a target rule  in
              confluence_store.(index) <- Some confluence;
              (a,leaf_store)))))
      targets
      (a,leaf_store)  in
  let a,leaf_store = fold entry_points a leaf_store  in
  a,(leaf_store,cycle_store,confluence_store)

let extract_subscopes tree =
  let rec do_it map subtree = List.fold_left
    (fun map -> fun node -> match node with
      | Oinmfile.Rule rule -> (match rule.Rule.drain with
        | Rule.Files _ | Rule.Scope None -> map
        | Rule.Scope (Some subtree) -> Rule.HalfruleMap.add
          rule.Rule.pre  subtree  (do_it map subtree))
      | Oinmfile.Generated targets -> map)
    map  subtree  in
  do_it Rule.HalfruleMap.empty tree

