(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Multiple_Rules of string

module Rulebase : Debugable_types.Type

type drain = Rule.Targets.t

val empty : Rulebase.t
val merge_bases : Rulebase.t -> Rulebase.t -> Rulebase.t
val base_of_tree : Oinmfile.tree -> Rulebase.t
val base_of_scope : Oinmfile.flat_scope -> Rulebase.t
val has_rule : Rulebase.t -> Rule.Target.t -> bool
val get_rule : Rulebase.t -> Rule.Target.t -> drain Rule.rule
val get_rule_optional : Rulebase.t -> Rule.Target.t -> drain Rule.rule option

type ('walk,'skip,'cycle,'confluence) subdigraph_options =
  | Enter of 'skip * 'cycle * 'walk
  | Ignore of 'cycle * 'confluence * 'walk

type ('leaf,'cycle,'confluence) fold_store
  (* A finished depth-first walk (see description of "fold_rulebase" below)
     may be continued with new entry points.     
     Then, the internal store (see below) needs to be preserved between
     invocations of fold_rulebase. *)

module Fold_store
  (Leaf : Value_types.Type)
  (Cycle : Value_types.Type)
  (Confluence : Value_types.Type) :
  Debugable_types.Type with type t = (Leaf.t, Cycle.t, Confluence.t) fold_store

val fold_rulebase :
      (* A depth-first walk through the dependency digraph,
        where leaves are targets for which there are no rules
        and all other nodes are rules.
        'walk is the data which is folded across the walk.
        'skip is for data that is not handed to subwalks.
        'cycle, 'confluence, and 'leaf are for data to be handed to the walk
        when it visits a node again. These are "stored" at the node.
	'cycle is intended for revisits that occur, while the depth-first walk
        is still below the node. However it is also available on other
        revisits. Note that this can only apply to non-leaf nodes.
        'confluence is intended for revisits of non-leaf nodes that occur after
	the depth-first walk has left the node.
        'leaf is intended for revisits of leaf nodes.
      *)
    ('walk -> Rule.Target.t -> 'leaf * 'walk) ->
        (* f_leaf *)
    ('leaf -> 'walk -> Rule.Target.t -> 'leaf * 'walk) ->
        (* f_revisit_leaf *)
    ('walk -> Rule.Target.t -> Rule.Targets.t Rule.rule ->
      ('walk,'skip,'cycle,'confluence) subdigraph_options) ->
        (* f_rule_start *)
    ('skip -> 'cycle -> 'walk -> Rule.Target.t -> Rule.Targets.t Rule.rule ->
      'confluence * 'walk) ->
        (* f_rule_end *)
    ('cycle -> 'walk -> Rule.Target.t -> Rule.Targets.t Rule.rule ->
      'cycle * 'walk) ->
        (* f_revisit_cycle *)
    ('cycle -> 'confluence -> 'walk ->
      Rule.Target.t -> Rule.Targets.t Rule.rule ->
      'confluence * 'walk) ->
        (* f_revisit_confluence *)
    ?store:('leaf,'cycle,'confluence) fold_store ->
      (* If this argument is given (which implies that it stems from the
         return value of a previous call to fold_rulebase (there is no other
         way to obtain a value of this type)), the depth-first walk is a
         continuation of the previous one in the sense that the nodes visited
         are the same as if there had been only one walk, which first used the
         previous entry points and now the new ones.
         Nevertheless, the actions performed at the various nodes are given by
         the values of f_leaf through f_revisit_confluence for the second
         (current) invocation.
         It is required that continuations work on extensions of the rulebase
         that the previous calls to fold_rulebase did. (`Extension' refers to
         (repeated) applications of (fun b -> merge_bases b somethingelse).)
         If this condition is not met, the result is unspecified. *)
    Rulebase.t -> 'walk ->
    Rule.Targets.t ->
      (* entry points *)
    'walk * ('leaf,'cycle,'confluence) fold_store

val extract_subscopes : Oinmfile.tree -> Oinmfile.tree Rule.HalfruleMap.pt
  (* Prepares all direct and indirect subscopes for search *)

