(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

let append_targets = List.fold_left
  (fun sofar -> fun target -> sofar^" "^target)

let add_suffix new_suffices bases = List.concat (List.map
  (fun base -> List.map ((^) base) new_suffices)
  bases)

let replace_suffix new_suffices deps = List.concat (List.map
  (fun dep ->
    let base = Fileinfo.kill_suffix dep Depsfile.suffix  in
    List.map (fun suf -> base^suf) new_suffices)
  deps)

let replace_suffix new_suffices deps = add_suffix new_suffices
  (List.map (fun dep -> Fileinfo.kill_suffix dep Depsfile.suffix) deps)


let promise targets = Oinmfile.FGenerated (Rule.TargetSet.from_list targets)


let compile_rule options job deps link_deps =
  let deps = if Job.has_source job
  then Job.source_name job :: deps
  else deps  in
  Oinmfile.FRule {
    Rule.pre = {
      Rule.sources = Rule.TargetSet.from_list deps;
      Rule.actions = [append_targets
        (Job.compile_command options job)
        link_deps];
      };
    Rule.drain = Rule.Files (Rule.TargetSet.from_list (
      Job.target_name job :: Job.extra_targets job));
    }


let depsfile_rules call options job deps_info =
  let depsfile_deps = List.map
    (fun info -> Fileinfo.base_name info ^ Depsfile.suffix)
    (List.filter
      (fun info -> Fileinfo.has_depsfile (Fileinfo.filetype info))
      deps_info)  in
  let depsfile_deps_set = Rule.TargetSet.from_list depsfile_deps  in
  let depsfile = Job.depsfile_name job  in
  let promises_rule = Oinmfile.FGenerated depsfile_deps_set  in
  let depsfile_rule = Oinmfile.FRule {
    Rule.pre = {
      Rule.sources = depsfile_deps_set;
      Rule.actions = [append_targets
        (call ^ Options.forward options ^ " " ^ depsfile)
        depsfile_deps]
      };
    Rule.drain = Rule.Files (Rule.TargetSet.singleton depsfile);
    }  in
  [promises_rule; depsfile_rule]


let rules_for_dependencies call options deps deps_info =
  let promises_rule = promise deps  in
  let compile_rules = List.map
    (fun dep_info -> Oinmfile.FRule {
      Rule.pre = {
        Rule.sources = Rule.TargetSet.singleton (Fileinfo.source_name dep_info);
        Rule.actions = [call ^ Options.forward options ^
          " " ^ Fileinfo.target_name dep_info];
        };
      Rule.drain = Rule.Scope ()
      })
    deps_info  in
  promises_rule :: compile_rules


let outer_to_inner_rule call options job =
  let has_source = Job.has_source job  in
  let sources = [Job.depsfile_name job]  in
  let sources = if has_source
  then Job.source_name job :: sources
  else sources  in
  Oinmfile.FRule {
    Rule.pre = {
      Rule.sources = Rule.TargetSet.from_list sources;
      Rule.actions = [call ^
        " " ^ Options.forward options ^
        " " ^ Job.outer_to_inner_option job ^
        " " ^ Job.target_name job ^
        if has_source  then (" " ^ Job.source_name job)  else ""];
      };
    Rule.drain = Rule.Scope ();
    }


let delay call options lexers parsers job =
  let generated_sources = Rule.TargetSet.from_list (
    add_suffix [".ml"] lexers @
    add_suffix [".ml"; ".mli"] parsers)  in
  Oinmfile.FGenerated generated_sources ::
  Job.reproduce call options generated_sources job ::
  List.map
    (fun lexer -> Oinmfile.FRule {
      Rule.pre = {
        Rule.sources = Rule.TargetSet.empty;
        Rule.actions = [call ^
          " " ^ Options.forward options ^
          " -lex " ^ lexer ^ ".mll"]
      };
      Rule.drain = Rule.Scope ();
    })
    lexers @
  List.map
    (fun parser -> Oinmfile.FRule {
      Rule.pre = {
        Rule.sources = Rule.TargetSet.empty;
        Rule.actions = [call ^
          " " ^ Options.forward options ^
          " -yacc " ^ parser ^ ".mly"]
      };
      Rule.drain = Rule.Scope ();
    })
    parsers


let gen_rules call options lexers parsers job =
  if lexers=[] & parsers=[]
  then match Job.mode job with
    | Mode.By_type Fileinfo.CMI ->
      let deps = Get_deps.get_deps job  in
      let deps_info = List.map Fileinfo.deduce deps  in
      compile_rule options job deps [] ::
      rules_for_dependencies call options deps deps_info
    | Mode.By_type Fileinfo.CMO | Mode.By_type Fileinfo.CMX ->
      let deps = Get_deps.get_deps job  in
      let deps_fixed = if Job.mode job = Mode.By_type Fileinfo.CMX
      then List.concat (List.map
        (fun dep -> if Filename.check_suffix dep ".cmx"
	  then [dep; Filename.chop_suffix dep ".cmx" ^ ".cmi"]
	  else [dep])
        deps)
      else deps  in
      let deps_info = List.map Fileinfo.deduce deps  in
      compile_rule options job deps_fixed [] ::
      (depsfile_rules call options job deps_info @
      rules_for_dependencies call options deps deps_info)
    | Mode.By_type Fileinfo.CMA | Mode.By_type Fileinfo.CMXA ->
      let deps = Job.given_deps job  in
      let deps_info = List.map Fileinfo.deduce deps  in
      outer_to_inner_rule call options job ::
      (depsfile_rules call options job deps_info @
      rules_for_dependencies call options deps deps_info)
    | Mode.Prog ->
      let deps = Get_deps.get_deps job  in
      let deps_info = List.map Fileinfo.deduce deps  in
      outer_to_inner_rule call options job ::
      (depsfile_rules call options job deps_info @
      rules_for_dependencies call options deps deps_info)
    | Mode.Prog_inner ->
      let raw_deps = Depsfile.read_deps (Job.depsfile_name job)  in
      let link_deps =
        options.Options.used_libs ::
        replace_suffix [".cmx"] raw_deps @ [Job.source_name job]  in
      let deps = replace_suffix [".cmx"; ".o"] raw_deps  in
      [compile_rule options job deps link_deps;
      promise deps]
    | Mode.CMA_inner ->
      let raw_deps = Depsfile.read_deps (Job.depsfile_name job)  in
      let deps = replace_suffix [".cma"] raw_deps  in
      [compile_rule options job deps deps;
        promise deps]
    | Mode.CMXA_inner ->
      let raw_deps = Depsfile.read_deps (Job.depsfile_name job)  in
      let link_deps =
        options.Options.used_libs ::
        replace_suffix [".cmx"] raw_deps  in
      let deps = replace_suffix [".cmx"; ".o"] raw_deps  in
      [compile_rule options job deps link_deps;
      promise deps]
    | Mode.Lex | Mode.Yacc -> [compile_rule options job [] []]
  else delay call options lexers parsers job


let out_rules scope = Oinmfile.output_tree stdout (Oinmfile.flat_to_tree scope)

