(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Not_deducible

type names = {
  source : string option;
  target : string option;
  ocamldep : string option;
  depsfile : string option;
  given_deps : string list option;
  extra_targets : string list option;
}

let bogus = {
  source = None;
  target = None;
  ocamldep = None;
  depsfile = None;
  given_deps = None;
  extra_targets = None;
}

type job = Mode.mode * names


let deduce target extra_args =
  let info = Fileinfo.deduce target  in
  let ft = Fileinfo.filetype info  in
  let source = match ft with
  | Fileinfo.CMI | Fileinfo.CMO | Fileinfo.CMX | Fileinfo.MLDEPS ->
    Some (Fileinfo.source_name info)
  | Fileinfo.CMA | Fileinfo.CMXA -> None  in
  let target_ = Some (Fileinfo.target_name info)  in
  let ocamldep = Some (Fileinfo.target_name info)  in
  let depsfile = match ft with
  | Fileinfo.CMO | Fileinfo.CMX | Fileinfo.MLDEPS ->
    Some (Fileinfo.base_name info ^ Depsfile.suffix)
  | Fileinfo.CMI -> None
  | Fileinfo.CMA | Fileinfo.CMXA -> Some (target ^ Depsfile.suffix)  in
  let given_deps = match ft with
  | Fileinfo.CMI | Fileinfo.CMO | Fileinfo.CMX -> None
  | Fileinfo.CMA | Fileinfo.CMXA | Fileinfo.MLDEPS -> Some extra_args  in
  let extra_targets = match ft with
  | Fileinfo.CMI | Fileinfo.CMO -> Some []
  | Fileinfo.CMX -> Some [Fileinfo.base_name info ^ ".o"]
  | Fileinfo.CMA | Fileinfo.CMXA | Fileinfo.MLDEPS -> None  in
  Mode.By_type (Fileinfo.filetype info), {bogus with
    source = source;
    target = target_;
    ocamldep = ocamldep;
    depsfile = depsfile;
    given_deps = given_deps;
    extra_targets = extra_targets;
  }

let program target source = Mode.Prog,
  let base = Fileinfo.kill_suffix source ".ml"  in
  {bogus with
    source = Some source;
    target = Some target;
    ocamldep = Some (base^".cmx");
    depsfile = Some (base^Depsfile.suffix);
  }

let program_inner target source = Mode.Prog_inner,
  let base = Fileinfo.kill_suffix source ".ml"  in
  {bogus with
    source = Some source;
    target = Some target;
    ocamldep = Some (base^".cmx");
    depsfile = Some (base^Depsfile.suffix);
    extra_targets = Some [];
  }

let cma_inner target = Mode.CMA_inner,
  {bogus with
    target = Some target;
    depsfile = Some (target^Depsfile.suffix);
    extra_targets = Some [];
  }

let cmxa_inner target = Mode.CMXA_inner,
  {bogus with
    target = Some target;
    depsfile = Some (target^Depsfile.suffix);
    extra_targets = Some [Fileinfo.base_name (Fileinfo.deduce target) ^ ".a"];
  }

let lex mllfile = Mode.Lex,
  let base = Fileinfo.kill_suffix mllfile ".mll"  in
  {bogus with
    source = Some mllfile;
    target = Some (base^".ml");
    extra_targets = Some [];
  }

let yacc mlyfile = Mode.Yacc,
  let base = Fileinfo.kill_suffix mlyfile ".mly"  in
  {bogus with
    source = Some mlyfile;
    target = Some (base^".ml");
    extra_targets = Some [base^".mli"; base^".output"];
  }


let mode (mode,names) = mode
let source_name (mode,names) = match names.source with Some s -> s
let target_name (mode,names) = match names.target with Some s -> s
let ocamldep_name (mode,names) = match names.ocamldep with Some s -> s
let depsfile_name (mode,names) = match names.depsfile with Some s -> s
let given_deps (mode,names) = match names.given_deps with Some ds -> ds
let extra_targets (mode,names) = match names.extra_targets with Some ts -> ts

let ocamlc_command options = "ocamlc.opt " ^ options.Options.ocamlc
let ocamlopt_command options = "ocamlopt.opt " ^ options.Options.ocamlopt

let compile_command options job =
  match mode job with
  | Mode.By_type Fileinfo.CMI | Mode.By_type Fileinfo.CMO ->
    ocamlc_command options ^ " -c " ^ source_name job
  | Mode.By_type Fileinfo.CMX ->
    ocamlopt_command options ^ " -c " ^ source_name job
  | Mode.Prog_inner -> ocamlopt_command options ^ " -o " ^ target_name job
  | Mode.CMA_inner -> ocamlc_command options ^ " -a -o " ^ target_name job
  | Mode.CMXA_inner -> ocamlopt_command options ^ " -a -o " ^ target_name job
  | Mode.Lex -> "ocamllex.opt " ^ source_name job
  | Mode.Yacc -> "ocamlyacc -v " ^ source_name job

let outer_to_inner_option job =
  Mode.mode_to_option (match mode job with
  | Mode.By_type Fileinfo.CMA -> Mode.CMA_inner
  | Mode.By_type Fileinfo.CMXA -> Mode.CMXA_inner
  | Mode.Prog -> Mode.Prog_inner)

let has_source job = match mode job with
  | Mode.By_type Fileinfo.CMI
  | Mode.By_type Fileinfo.CMO | Mode.By_type Fileinfo.CMX -> true
  | Mode.By_type Fileinfo.CMA | Mode.By_type Fileinfo.CMXA -> false
  | Mode.Prog | Mode.Prog_inner -> true
  | Mode.CMA_inner | Mode.CMXA_inner -> false
  | Mode.Lex | Mode.Yacc -> true

let oinm4ocaml_deps job = match mode job with
  | Mode.By_type Fileinfo.CMI
  | Mode.By_type Fileinfo.CMO
  | Mode.By_type Fileinfo.CMX
  | Mode.Prog
    -> [source_name job]
  | Mode.By_type Fileinfo.CMA
  | Mode.By_type Fileinfo.CMXA
    -> []
  | Mode.Prog_inner
  | Mode.CMA_inner
  | Mode.CMXA_inner
    -> [depsfile_name job]
  | Mode.By_type Fileinfo.MLDEPS
    -> given_deps job

let reproduce call options extra_deps job =
  let mode = mode job  in
  let main_call = match mode with
  | Mode.By_type Fileinfo.CMI
  | Mode.By_type Fileinfo.CMO | Mode.By_type Fileinfo.CMX -> target_name job
  | Mode.By_type Fileinfo.CMA | Mode.By_type Fileinfo.CMXA
  | Mode.By_type Fileinfo.MLDEPS -> List.fold_left
    (fun so_far -> fun gd -> so_far ^ " " ^ gd)
    (target_name job)
    (given_deps job)
  | Mode.Prog | Mode.Prog_inner ->
    Mode.mode_to_option mode ^ " " ^ (target_name job) ^ " " ^ (source_name job)
  | Mode.CMA_inner | Mode.CMXA_inner | Mode.Lex | Mode.Yacc ->
    Mode.mode_to_option mode ^ " " ^ (target_name job)  in
  Oinmfile.FRule {
    Rule.pre = {
      Rule.sources = Rule.TargetSet.union extra_deps
        (Rule.TargetSet.from_list (oinm4ocaml_deps job));
      Rule.actions = [call ^ Options.forward options ^ " " ^ main_call]
    };
    Rule.drain = Rule.Scope ();
  }
