(*
   Copyright 2007-2012 by Mark Weyer
   Licensed under the OINM licence version 0
*)

exception Wrong_file_format

let status_file_name = ".oinm.status"
let file_format_version = 1


module Hash_Collection = Rule.TargetMap.Specialized(Hash.Hash)

let get_collection cache targets = Rule.TargetSet.fold
  (fun target -> fun collection -> Rule.TargetMap.add
    target (Hash.get_hash cache target) collection)
  targets
  Rule.TargetMap.empty

let up_to_date cache collection = Rule.TargetMap.fold
  (fun target -> fun hash -> fun unchanged_so_far ->
    unchanged_so_far && 0=Hash.Hash.compare hash (Hash.get_hash cache target))
  collection
  true


module Local_Status = Value_types.Pair(Hash_Collection)(Hash_Collection)

module UDCR = Rule.Rule.Specialized
  (Rule.Drain_cases.Specialized(Value_types.Unit))
module UDCR_Map = Value_types.Map(UDCR)
module Global_Status = UDCR_Map.Specialized(Local_Status)

let pre_rule cache global rule =
  let source_hashes = get_collection cache rule.Rule.pre.Rule.sources  in
  (match UDCR_Map.find_optional rule global  with
  | None -> true
  | Some (old_source_hashes,old_drain_hashes) ->
    (Hash_Collection.compare source_hashes old_source_hashes <> 0)
    || not (up_to_date cache old_drain_hashes)),
  source_hashes

let post_rule cache global rule source_hashes =
  let drain_hashes = match rule.Rule.drain with
  | Rule.Scope () -> Rule.TargetMap.empty
  | Rule.Files drains -> (
    Rule.TargetSet.iter (fun target -> Hash.uncache target cache) drains;
    get_collection cache drains)  in
  UDCR_Map.add rule (source_hashes, drain_hashes) global

let save_global_status global =
  let file = open_out status_file_name  in
  Value_types.Int.save file file_format_version;
  Global_Status.save file global;
  close_out file

let empty_global = UDCR_Map.empty

let load_global_status () =
  try
    let file = open_in status_file_name  in
    let global = try
      if Value_types.Int.load file != file_format_version
      then raise Wrong_file_format;
      Global_Status.load file
    with
    | Value_types.Malformed_input -> (
      prerr_string ("Warning: File "^status_file_name^" corrupted.\n"^
        "Using empty file instead.\n");
      flush stderr;
      empty_global)  in
    close_in file;
    global
  with
  | Sys_error s -> (
    prerr_string ("Warning: File "^status_file_name^" could not be opened.\n"^
      "Error message was \""^s^"\".\n"^
      "Using empty file instead.\n");
    flush stderr;
    empty_global)

